"""
The View classes render the data returned by a Source as a Panel
object.
"""

import sys

from io import StringIO
from weakref import WeakKeyDictionary

import numpy as np
import param
import panel as pn

from bokeh.models import NumeralTickFormatter
from panel.pane.base import PaneBase
from panel.pane.perspective import (
    THEMES as _PERSPECTIVE_THEMES, Plugin as _PerspectivePlugin
)
from panel.param import Param

from ..config import _INDICATORS
from ..filters import ParamFilter
from ..sources import Source
from ..transforms import Transform
from ..util import resolve_module_reference


class View(param.Parameterized):
    """
    A View renders the data returned by a Source as a Viewable Panel
    object. The base class provides methods which query the Source for
    the latest data given the current filters and applies all
    specified `transforms`.

    Subclasses should use these methods to query the data and return
    a Viewable Panel object in the `get_panel` method.
    """

    controls = param.List(default=[], doc="""
        Parameters that should be exposed as widgets in the UI.""")

    filters = param.List(constant=True, doc="""
        A list of Filter object providing the query parameters for the
        Source.""")

    source = param.ClassSelector(class_=Source, constant=True, doc="""
        The Source to query for the data.""")

    selection_group = param.String(default=None, doc="""
        Declares a selection group the plot is part of. This feature
        requires the separate HoloViews library.""")

    transforms = param.List(constant=True, doc="""
        A list of transforms to apply to the data returned by the
        Source before visualizing it.""")

    table = param.String(doc="The table being visualized.")

    field = param.Selector(doc="The field being visualized.")

    view_type = None

    # Panel extension to load to render this View
    _extension = None

    # Parameters which reference fields in the table
    _field_params = ['field']

    _selections = WeakKeyDictionary()

    _supports_selections = False

    __abstract = True

    def __init__(self, **params):
        self._cache = None
        self._ls = None
        self._panel = None
        self._updates = None
        self.kwargs = {k: v for k, v in params.items() if k not in self.param}

        # Populate field selector parameters
        params = {k: v for k, v in params.items() if k in self.param}
        source, table = params.pop('source', None), params.pop('table', None)
        if source is None:
            raise ValueError("Views must declare a Source.")
        if table is None:
            raise ValueError("Views must reference a table on the declared Source.")
        fields = list(source.get_schema(table))
        for fp in self._field_params:
            if isinstance(self.param[fp], param.ObjectSelector):
                self.param[fp].objects = fields

        super().__init__(source=source, table=table, **params)
        if self.selection_group:
            self._init_link_selections()

    def _init_link_selections(self):
        if self._ls is not None:
            return
        doc = pn.state.curdoc
        if doc not in View._selections and self.selection_group:
            View._selections[doc] = {}
        self._ls = View._selections.get(doc, {}).get(self.selection_group)
        if self._ls is None:
            from holoviews.selection import link_selections
            self._ls = link_selections.instance()
            if self.selection_group:
                View._selections[doc][self.selection_group] = self._ls
        if 'selection_expr' in self.param:
            self._ls.param.watch(self._update_selection_expr, 'selection_expr')

    def _update_selection_expr(self, event):
        self.selection_expr = event.new

    @classmethod
    def _get_type(cls, view_type):
        """
        Returns the matching View type.
        """
        if '.' in view_type:
            return resolve_module_reference(view_type, View)
        try:
            __import__(f'lumen.views.{view_type}')
        except Exception:
            pass
        for view in param.concrete_descendents(cls).values():
            if view.view_type == view_type:
                return view
        if view_type is not None:
            raise ValueError(f"View type '{view_type}' could not be found.")
        return View

    @classmethod
    def from_spec(cls, spec, source, filters):
        """
        Resolves a View specification given the schema of the Source
        it will be filtering on.

        Parameters
        ----------
        spec: dict
            Specification declared as a dictionary of parameter values.
        source: lumen.sources.Source
            The Source object containing the tables the View renders.
        filters: list(lumen.filters.Filter)
            A list of Filter objects which provide query values for
            the Source.

        Returns
        -------
        The resolved View object.
        """
        spec = dict(spec)
        transform_specs = spec.pop('transforms', [])
        transforms = [Transform.from_spec(tspec) for tspec in transform_specs]
        view_type = View._get_type(spec.pop('type', None))
        resolved_spec = {}
        for p, value in spec.items():
            if p not in view_type.param:
                resolved_spec[p] = value
                continue
            parameter = view_type.param[p]
            if isinstance(parameter, param.ObjectSelector) and parameter.names:
                try:
                    value = parameter.names.get(value, value)
                except Exception:
                    pass
            resolved_spec[p] = value
        view = view_type(
            filters=filters, source=source, transforms=transforms,
            **resolved_spec
        )

        # Resolve ParamFilter parameters
        for filt in filters:
            if isinstance(filt, ParamFilter):
                if not isinstance(filt.parameter, str):
                    continue
                name, parameter = filt.parameter.split('.')
                if name == view.name and parameter in view.param:
                    filt.parameter = view.param[parameter]
        return view

    def __bool__(self):
        return self._cache is not None and len(self._cache) > 0

    def _update_panel(self, *events):
        """
        Updates the cached Panel object and returns a boolean value
        indicating whether a rerender is required.
        """
        if self._panel is not None:
            self._cleanup()
            self._updates = self._get_params()
            if self._updates is not None:
                return False
        self._panel = self.get_panel()
        return True

    def _cleanup(self):
        """
        Method that is called on update.
        """

    def get_data(self):
        """
        Queries the Source for the specified table applying any
        filters and transformations specified on the View. Unlike
        `get_value` this should be used when multiple return values
        are expected.

        Returns
        -------
        DataFrame
            The queried table after filtering and transformations are
            applied.
        """
        if self._cache is not None:
            return self._cache
        query = {
            filt.field: filt.query for filt in self.filters
            if filt.query is not None and
            (filt.table is None or filt.table == self.table)
        }
        data = self.source.get(self.table, **query)
        for transform in self.transforms:
            data = transform.apply(data)
        if len(data):
            data = self.source._filter_dataframe(data, **query)
        for filt in self.filters:
            if not isinstance(filt, ParamFilter):
                continue
            from holoviews import Dataset
            if filt.value is not None:
                ds = Dataset(data)
                data = ds.select(filt.value).data
        self._cache = data
        return data

    def get_value(self, field=None):
        """
        Queries the Source for the data associated with a particular
        field applying any filters and transformations specified on
        the View. Unlike `get_data` this method returns a single
        scalar value associated with the field and should therefore
        only be used if only a single.

        Parameters
        ----------
        field: str (optional)
            The field from the table to return; if None uses
            field defined on the View.

        Returns
        -------
        object
            A single scalar value representing the current value of
            the queried field.
        """
        data = self.get_data()
        if not len(data) or field is not None and field not in data.columns:
            return None
        row = data.iloc[-1]
        return row[self.field if field is None else field]

    def get_panel(self):
        """
        Constructs and returns a Panel object which will represent a
        view of the queried table.

        Returns
        -------
        panel.Viewable
            A Panel Viewable object representing a current
            representation of the queried table.
        """
        return pn.panel(self.get_data())

    def update(self, *events, invalidate_cache=True):
        """
        Triggers an update in the View.

        Parameters
        ----------
        events: tuple
            param events that may trigger an update.
        invalidate_cache : bool
            Whether to clear the View's cache.

        Returns
        -------
        stale : bool
            Whether the panel on the View is stale and needs to be
            rerendered.
        """
        if invalidate_cache:
            self._cache = None
        return self._update_panel()

    def _get_params(self):
        return None

    @property
    def control_panel(self):
        return Param(self.param, parameters=self.controls, sizing_mode='stretch_width')

    @property
    def panel(self):
        if isinstance(self._panel, PaneBase):
            pane = self._panel
            if len(pane.layout) == 1 and pane._unpack:
                return pane.layout[0]
            return pane._layout
        return self._panel


class StringView(View):
    """
    The StringView renders the latest value of the field as a HTML
    string with a specified fontsize.
    """

    font_size = param.String(default='24pt', doc="""
        The font size of the rendered field value.""")

    view_type = 'string'

    def get_panel(self):
        return pn.pane.HTML(**self._get_params())

    def _get_params(self):
        value = self.get_value()
        params = dict(self.kwargs)
        if value is None:
            params['object'] = 'No info'
        else:
            params['object'] = f'<p style="font-size: {self.font_size}">{value}</p>'
        return params



class IndicatorView(View):
    """
    The IndicatorView renders the latest field value as a Panel
    Indicator.
    """

    indicator = param.Selector(objects=_INDICATORS, doc="""
        The name of the panel Indicator type.""")

    label = param.String(doc="""
        A custom label to use for the Indicator.""")

    view_type = 'indicator'

    def __init__(self, **params):
        super().__init__(**params)
        name = params.get('label', params.get('field', ''))
        self.kwargs['name'] = name

    def get_panel(self):
        return self.indicator(**self._get_params())

    def _get_params(self):
        params = dict(self.kwargs)
        if 'data' in self.indicator.param:
            params['data'] = self.get_data()
        else:
            value = self.get_value()
            if (not isinstance(value, (type(None), str)) and np.isnan(value)):
                value = None
            params['value'] = value
        return params


class hvPlotView(View):
    """
    The hvPlotView renders the queried data as a bokeh plot generated
    with hvPlot. hvPlot allows for a concise declaration of a plot via
    its simple API.
    """

    kind = param.String(doc="The kind of plot, e.g. 'scatter' or 'line'.")

    x = param.Selector(doc="The column to render on the x-axis.")

    y = param.Selector(doc="The column to render on the y-axis.")

    by = param.ListSelector(doc="The column(s) to facet the plot by.")

    groupby = param.ListSelector(doc="The column(s) to group by.")

    opts = param.Dict(default={}, doc="HoloViews options to apply on the plot.")

    streaming = param.Boolean(default=False, doc="""
        Whether to stream new data to the plot or rerender the plot.""")

    selection_expr = param.Parameter(doc="""
        A selection expression caputirng the current selection applied
        on the plot.""")

    view_type = 'hvplot'

    _field_params = ['x', 'y', 'by', 'groupby']

    _supports_selections = True

    def __init__(self, **params):
        import hvplot.pandas # noqa
        if 'dask' in sys.modules:
            try:
                import hvplot.dask # noqa
            except Exception:
                pass
        if 'by' in params and isinstance(params['by'], str):
            params['by'] = [params['by']]
        if 'groupby' in params and isinstance(params['groupby'], str):
            params['groupby'] = [params['groupby']]
        self._stream = None
        self._linked_objs = []
        super().__init__(**params)

    def get_plot(self, df):
        processed = {}
        for k, v in self.kwargs.items():
            if k.endswith('formatter') and isinstance(v, str) and '%' not in v:
                v = NumeralTickFormatter(format=v)
            processed[k] = v
        if self.streaming:
            processed['stream'] = self._stream
        plot = df.hvplot(
            kind=self.kind, x=self.x, y=self.y, **processed
        )
        plot = plot.opts(**self.opts) if self.opts else plot
        if self.selection_group or 'selection_expr' in self._param_watchers:
            plot = self._link_plot(plot)
        return plot

    def _link_plot(self, plot):
        self._init_link_selections()
        linked_objs = list(self._ls._plot_reset_streams)
        plot = self._ls(plot)
        self._linked_objs += [
            o for o in self._ls._plot_reset_streams if o not in linked_objs
        ]
        return plot

    def _cleanup(self):
        if self._ls is None:
            return
        for obj in self._linked_objs:
            reset = self._ls._plot_reset_streams.pop(obj)
            sel_expr = self._ls._selection_expr_streams.pop(obj)
            self._ls._cross_filter_stream.input_streams.remove(sel_expr)
            sel_expr.clear()
            sel_expr.source = None
            reset.clear()
            reset.source = None
        self._linked_objs = []

    def get_panel(self):
        return pn.pane.HoloViews(**self._get_params())

    def _get_params(self):
        df = self.get_data()
        if self.streaming:
            from holoviews.streams import Pipe
            self._stream = Pipe(data=df)
        return dict(object=self.get_plot(df))

    def update(self, *events, invalidate_cache=True):
        """
        Triggers an update in the View.

        Parameters
        ----------
        events: tuple
            param events that may trigger an update.
        invalidate_cache : bool
            Whether to clear the View's cache.

        Returns
        -------
        stale : bool
            Whether the panel on the View is stale and needs to be
            rerendered.
        """
        # Skip events triggered by a parameter change on this View
        own_parameters = [self.param[p] for p in self.param]
        own_events = events and all(
            isinstance(e.obj, ParamFilter) and
            (e.obj.parameter in own_parameters or
            e.new is self._ls.selection_expr)
            for e in events
        )
        if own_events:
            return False
        if invalidate_cache:
            self._cache = None
        if not self.streaming or self._stream is None:
            return self._update_panel()
        self._stream.send(self.get_data())
        return False


class Table(View):
    """
    Renders a Source table using a Panel Table widget.
    """

    view_type = 'table'

    def get_panel(self):
        return pn.widgets.tables.Tabulator(**self._get_params())

    def _get_params(self):
        return dict(value=self.get_data(), disabled=True, **self.kwargs)


class Download(View):
    """
    The Download View allows downloading the current table as a csv or
    xlsx file.
    """

    filename = param.String(default='data', doc="""
      Filename of the downloaded file.""")

    filetype = param.Selector(default='csv', objects=['csv', 'xlsx'], doc="""
      File type of the downloaded file.""")

    save_kwargs = param.Dict(default={}, doc="""
      Options for the to_csv or to_excel methods.""")

    view_type = 'download'

    def __bool__(self):
        return True

    def _get_file_data(self):
        df = self.get_data()
        sio = StringIO()
        if self.filetype == 'csv':
            savefn = df.to_csv
        elif self.filetype == 'xlsx':
            savefn = df.to_excel
        savefn(sio, **self.save_kwargs)
        sio.seek(0)
        return sio

    def get_panel(self):
        return pn.widgets.FileDownload(**self._get_params())

    def _get_params(self):
        filename = f'{self.filename}.{self.filetype}'
        return dict(filename=filename, callback=self._get_file_data, **self.kwargs)



class PerspectiveView(View):

    aggregates = param.Dict(None, doc="""
        How to aggregate. For example {x: "distinct count"}""")

    columns = param.ListSelector(default=None, doc="""
        A list of source columns to show as columns. For example ["x", "y"]""")

    computed_columns = param.ListSelector(default=None, doc="""
        A list of computed columns. For example [""x"+"index""]""")

    column_pivots = param.ListSelector(None, doc="""
        A list of source columns to pivot by. For example ["x", "y"]""")

    filters = param.List(default=None, doc="""
        How to filter. For example [["x", "<", 3],["y", "contains", "abc"]]""")

    row_pivots = param.ListSelector(default=None, doc="""
        A list of source columns to group by. For example ["x", "y"]""")

    selectable = param.Boolean(default=True, allow_None=True, doc="""
        Whether items are selectable.""")

    sort = param.List(default=None, doc="""
        How to sort. For example[["x","desc"]]""")

    plugin = param.ObjectSelector(default=_PerspectivePlugin.GRID.value, objects=_PerspectivePlugin.options(), doc="""
        The name of a plugin to display the data. For example hypergrid or d3_xy_scatter.""")

    theme = param.ObjectSelector(default='material', objects=_PERSPECTIVE_THEMES, doc="""
        The style of the PerspectiveViewer. For example material-dark""")

    view_type = 'perspective'

    _extension = 'perspective'

    _field_params = ['columns', 'computed_columns', 'column_pivots', 'row_pivots']

    def _get_params(self):
        df = self.get_data()
        param_values = dict(self.param.get_param_values())
        params = set(View.param) ^ set(PerspectiveView.param)
        kwargs = dict({p: param_values[p] for p in params}, **self.kwargs)
        return dict(object=df, toggle_config=False, **kwargs)

    def get_panel(self):
        return pn.pane.Perspective(**self._get_params())


class AltairView(View):

    chart = param.Dict(default={}, doc="Keyword argument for Chart.")

    x = param.Selector(doc="The column to render on the x-axis.")

    y = param.Selector(doc="The column to render on the y-axis.")

    marker = param.Selector(default='line', objects=[
        'area', 'bar', 'boxplot', 'circle', 'errorband', 'errorbar',
        'geoshape', 'image', 'line', 'point', 'rect', 'rule', 'square',
        'text', 'tick', 'trail'])

    encode = param.Dict(default={}, doc="Keyword arguments for encode.")

    mark = param.Dict(default={}, doc="Keyword arguments for mark.")

    transform = param.Dict(doc="""
        Keyword arguments for transforms, nested by the type of
        transform, e.g. {'bin': {'as_': 'binned', 'field': 'x'}}.""")

    project = param.Dict(doc="Keyword arguments for project.")

    properties = param.Dict(doc="Keyword arguments for properties.")

    view_type = 'altair'

    _extension = 'vega'

    def _transform_encoding(self, encoding, value):
        import altair as alt
        if isinstance(value, dict):
            value = dict(value)
            for kw, val in value.items():
                if kw == 'scale':
                    if isinstance(val, list):
                        val = alt.Scale(range=val)
                    else:
                        val = alt.Scale(**val)
                if kw == 'tooltip':
                    val = [alt.Tooltip(**v) for v in val]
                value[kw] = val
            value = getattr(alt, encoding.capitalize())(**value)
        return value

    def _get_params(self):
        import altair as alt
        df = self.get_data()
        chart = alt.Chart(df, **self.chart)
        mark = getattr(chart, f'mark_{self.marker}')(**self.mark)
        x = self._transform_encoding('x', self.x)
        y = self._transform_encoding('y', self.y)
        encode = {k: self._transform_encoding(k, v) for k, v in self.encode.items()}
        encoded = mark.encode(x=x, y=y, **encode)
        if self.transform:
            for key, kwargs in self.transform.items():
                encoded = getattr(encoded, f'transform_{key}')(**kwargs)
        if self.project:
            encoded = encoded.project(**self.project)
        if self.properties:
            encoded = encoded.properties(**self.properties)
        return dict(object=encoded, **self.kwargs)

    def get_panel(self):
        return pn.pane.Vega(**self._get_params())



import os
import pathlib
import uuid
import tempfile
import yaml

import panel as pn
import param

from panel.layout.base import ListLike
from panel.template.base import BasicTemplate
from panel.template import FastListTemplate

from lumen.config import config
from lumen.sources import Source
from lumen.state import state as lm_state

from .base import Wizard, WizardItem
from .config import ConfigEditor
from .gallery import Gallery, GalleryItem
from .sources import SourceGallery
from .state import state
from .targets import TargetGallery, TargetEditor, TargetGalleryItem
from .views import ViewEditor, ViewGallery, ViewGalleryItem

CSS = [
    'https://cdnjs.cloudflare.com/ajax/libs/font-awesome/5.15.3/css/all.min.css'
]

CSS_RAW = """
i.fa.fa-plus:hover {
  color: var(--neutral-fill-color) !important;
}

.gallery-item:hover {
  box-shadow: 0 1px 5px;
}

.gallery-item {
  cursor: pointer;
}
"""

pn.extension('ace', 'perspective', 'tabulator', css_files=CSS, raw_css=[CSS_RAW])


class DashboardGalleryItem(GalleryItem):

    _template = """
    <div id="dashboard-item" onclick="${_launch}">
      <span style="font-size: 1.25em; font-weight: bold;">{{ name }}</span>
      <div id="details" style="margin-top: 1em;">
        ${view}
      </div>
      <p style="height: 4em;">${description}</p>
    </div>
    """

    _thumbnail = pathlib.Path(__file__).parent / 'assets' / 'dashboard.png'

    def __init__(self, **params):
        super().__init__(**params)
        metadata = self.spec.get('metadata', {})
        thumbnail = metadata.get('thumbnail', self._thumbnail)
        self.description = metadata.get('description')
        self.view = pn.pane.PNG(str(thumbnail), height=150, margin=0, align='center')

    def _launch(self, event):
        self.selected = True



class DashboardGallery(WizardItem, Gallery):
    "Welcome to the Lumen Dashboard Builder!"

    auto_advance = param.Boolean(default=True)

    path = param.Foldername()

    spec = param.Dict(default={}, precedence=-1)

    _template = """
    <span style="font-size: 2em;">{{ __doc__ }}</span>
    <fast-divider style="margin: 1em 0;"></fast-divider>
    <span style="font-size: 1.5em">Select an existing dashboard, start from scratch or upload a dashboard to edit.</span>
    <fast-button id="upload" onclick="${_open_modal}" appearance="outline" style="float: right;">
        <b style="font-size: 1.5em;">+</b>
    </fast-button>
    <div id="items" style="margin: 1em 0; display: flex; flex-wrap: wrap; gap: 1em;">
    {% for item in items.values() %}
      <fast-card id="dashboard-container" class="gallery-item" style="height: 290px; width: 350px; padding: 1em;">
        ${item}
      </fast-card>
    {% endfor %}
      <fast-card id="dashboard-container-new" class="gallery-item" style="height: 290px; width: 350px; padding: 1em;">
        <div style="display: grid;">
          <span style="font-size: 1.25em; font-weight: bold;">Create new dashboard</span>
          <i id="add-button" onclick="${_create_new}" class="fa fa-plus" style="font-size: 14em; margin: 0.2em auto;" aria-hidden="true"></i>
        </div>
      </fast-card>
    </div>
    """

    _gallery_item = DashboardGalleryItem

    def __init__(self, **params):
        super().__init__(**params)
        self.editor = pn.widgets.Ace(
            language='yml', sizing_mode='stretch_both', theme='dracula',
            margin=(0, 20), height=400
        )
        self.upload = pn.widgets.FileInput(
            sizing_mode='stretch_width', margin=(0, 20), height=50
        )
        self.thumbnail_upload = pn.widgets.FileInput(
            sizing_mode='stretch_width', margin=0, name='Upload a thumbnail'
        )
        self.upload.param.watch(self._yaml_upload, 'value')
        self.namer = pn.widgets.TextInput(
            name='Dashboard name', placeholder='Give the dashboard a name',
            margin=(5, 0)
        )
        self.description = pn.widgets.TextAreaInput(
            name='Provide a description', sizing_mode='stretch_both',
            max_length=100
        )
        self.save = pn.widgets.Button(name='Save', sizing_mode='stretch_width')
        self.save.on_click(self._save)
        self._modal_content = [
            '# Upload dashboard\nUpload a yaml file',
            self.upload,
            pn.Row(
                pn.Column(self.namer, self.thumbnail_upload),
                self.description, sizing_mode='stretch_width',
                margin=(0, 20)
            ),
            self.editor,
            self.save
        ]

    def _save(self, event):
        name = self.namer.value
        spec = yaml.safe_load(self.editor.value)
        spec['metadata'] = {
            'description': self.description.value
        }
        if self.thumbnail_upload.value:
            ext = self.thumbnail_upload.filename.split('.')[-1]
            thumbnail_path = pathlib.Path(self.path) / f'{name}.{ext}'
            with open(thumbnail_path, 'wb') as f:
                f.write(self.thumbnail_upload.value)
            spec['metadata']['thumbnail'] = str(thumbnail_path)
        with open(pathlib.Path(self.path) / f'{name}.yml', 'w', encoding='utf-8') as f:
            f.write(yaml.dump(spec))
        self.items[name] = self._gallery_item(name=name, spec=spec)
        self.param.trigger('items')
        state.template.close_modal()
        self._reset_form()

    def _reset_form(self):
        self.namer.value = ''
        self.editor.value = ''
        self.description.value = ''
        self.upload.filename = ''
        self.upload.value = None
        self.thumbnail_upload.filename = ''
        self.thumbnail_upload.value = None

    def _yaml_upload(self, event):
        if event.new is None:
            return
        name = '.'.join(self.upload.filename.split('.')[:-1])
        self.namer.value = name
        self.editor.value = event.new.decode('utf-8')
        return

    def _create_new(self, event):
        self.spec = {'config': {}, 'sources': {}, 'targets': []}

    def _selected(self, event):
        self.spec = event.obj.spec


class Launcher(WizardItem):
    "WIP: Plugins will allow launching to various cloud providers."

    _template = """
    <span style="font-size: 1.5em">Launcher</span>
    <p>{{ __doc__ }}</p>
    <fast-divider></fast-divider>
    <div style="width: 100%">
      <fast-button id="launch-button" style="margin: 2em auto;" onclick="${_launch}">Launch</fast-button>
    </div>
    """

    def _launch(self, event):
        from lumen import Dashboard
        with tempfile.NamedTemporaryFile('w', suffix='.yaml', delete=False) as f:
            f.file.write(yaml.dump(state.spec))
        dashboard = Dashboard(f.name)
        dashboard._render_dashboard()
        dashboard.show()


class YAML(WizardItem):
    """
    Download or copy the dashboard yaml specification.
    """

    editor = param.ClassSelector(class_=pn.widgets.Ace)

    _template = """
    <span style="font-size: 1.5em">Launcher</span>
    <p>{{ __doc__ }}</p>
    <fast-divider></fast-divider>
    <div id="yaml">${editor}</div>
    """

    def __init__(self, spec, **params):
        yaml_text = yaml.dump(spec)
        params['editor'] = pn.widgets.Ace(value=yaml_text,
            language='yaml', sizing_mode='stretch_both', theme='dracula',
            margin=0
        )
        super().__init__(spec=spec, **params)

    @param.depends('active', watch=True)
    def _update_spec(self):
        self.spec = state.spec
        self.editor.value = yaml.dump(state.spec)


class Builder(param.Parameterized):

    component_dir = param.Path(default=None)

    spec = param.Dict(default={})

    template = param.ClassSelector(class_=BasicTemplate)

    modal = param.ClassSelector(class_=ListLike)

    def __init__(self, **params):
        path = params['component_dir']
        dash_params, source_params, target_params, view_params = {}, {}, {}, {}
        dash_params['path'] = os.path.join(path, 'dashboards')
        source_params['path'] = os.path.join(path, 'sources')
        target_params['path'] = os.path.join(path, 'targets')
        view_params['path'] = os.path.join(path, 'views')
        self.welcome = DashboardGallery(**dash_params)
        super().__init__(**params)

        self.config = ConfigEditor(spec=self.spec['config'])
        state.spec = self.spec
        state.sources = self.sources = SourceGallery(spec=self.spec['sources'], **source_params)
        state.views = self.views = ViewGallery(**view_params)
        state.targets = self.targets = TargetGallery(spec=self.spec['targets'], **target_params)
        self.launcher = YAML(spec=self.spec)
        self.wizard = Wizard(items=[
            self.welcome, self.config, self.sources, self.views, self.targets, self.launcher
        ], sizing_mode='stretch_both')

        preview = pn.widgets.Button(name='Preview', width=100)
        preview.on_click(self._open_dialog)

        state.template.header.append(preview)
        if pn.state.user:
            logout = pn.widgets.Button(
                name='✖', width=40, css_classes=['logout'], margin=0,
                align='center'
            )
            logout.js_on_click(code="""
            window.location.href = '/logout'
            """)
            self.template.header.extend([
                pn.layout.HSpacer(),
                f'<b><font size="4.5em">User: {pn.state.user}</font></b>',
                logout
            ])
        self.template.main.append(self.wizard)
        self.template.modal.append(self.modal)

    def _open_dialog(self, event):
        self.wizard.preview.object = dict(self.spec)
        self.wizard.open_modal()

    @param.depends('welcome.spec', watch=True)
    def _update_spec(self):
        self.wizard.loading = True
        for key, item in self.welcome.spec.items():
            if isinstance(item, list):
                self.spec[key][:] = item
            else:
                if key in self.spec:
                    self.spec[key].clear()
                    self.spec[key].update(item)
                else:
                    self.spec[key] = item
        self.config.param.trigger('spec')
        self.sources.param.trigger('spec')

        for name, source in self.sources.sources.items():
            config.root = str(pathlib.Path(__file__).parent)
            spec = source.spec
            if isinstance(source.spec, dict):
                spec = dict(source.spec)
                spec.pop('filters', None)
            lm_state.sources[name] = Source.from_spec(spec)

        views, editors = [], {}
        targets, target_items = [], {}
        for target in self.spec['targets']:
            source_spec = target['source']
            if isinstance(source_spec, dict):
                source_spec = dict(source_spec, cache_dir=None)
                source_spec.pop('filters', None)
            source = Source.from_spec(source_spec)
            view_specs = target['views']
            if isinstance(view_specs, list):
                view_specs = [(f"{view['type']}: {view['table']}", view) for view in view_specs]
            else:
                view_specs = view_specs.items()
            for name, view in view_specs:
                view = dict(view)
                view_type = view.pop('type')
                editor = ViewEditor(
                    view_type=view_type, name=name, source_obj=source,
                    spec=view
                )
                item = ViewGalleryItem(
                    editor=editor, name=name, selected=True
                )
                while name in editors:
                    editors[name] = item
                views.append(editor)

            editor = TargetEditor(spec=target, views=list(editors))
            item = TargetGalleryItem(spec=target, editor=editor)
            targets.append(editor)
            target_items[f'target{uuid.uuid4().hex}'] = item

        self.targets.param.set_param(targets=targets, items=target_items)
        self.views.param.set_param(views=views, items=editors)
        self.welcome.ready = True
        self.wizard.loading = False
        self.launcher.value = yaml.dump(self.spec)

    def servable(self, title='Lumen Builder'):
        self.template.servable(title=title)


def main():
    path = pathlib.Path('./components').absolute()
    path.mkdir(exist_ok=True)
    params = {'component_dir': str(path)}
    (path / 'dashboards').mkdir(parents=True, exist_ok=True)
    (path / 'sources').mkdir(parents=True, exist_ok=True)
    (path / 'targets').mkdir(parents=True, exist_ok=True)
    (path / 'views').mkdir(parents=True, exist_ok=True)
    state.modal = pn.Column(sizing_mode='stretch_both')
    state.spec = {'config': {}, 'sources': {}, 'targets': []}
    state.template = FastListTemplate(theme='dark', title='Lumen Builder')
    builder = Builder(template=state.template, spec=state.spec, modal=state.modal, **params)
    builder.servable()


if __name__.startswith('bokeh'):
    main()



import param

from panel.io.server import init_doc, state
from panel.layout.base import ListLike, NamedListLike
from panel.reactive import ReactiveHTML
from panel.widgets.select import SelectBase


class FastDesignProvider(ListLike, ReactiveHTML):

    _template = '<fast-design-system-provider id="fdsp" use-defaults>${objects}</fast-design-system-provider>'


class FastComponent(ReactiveHTML):

    sizing_mode = param.ObjectSelector(default='stretch_width', objects=[
        'fixed', 'stretch_width', 'stretch_height', 'stretch_both',
        'scale_width', 'scale_height', 'scale_both', None])

    __abstract = True

    def get_root(self, doc=None, comm=None, preprocess=True):
        doc = init_doc(doc)
        root_obj = FastDesignProvider()
        root_obj.append(self)
        root = root_obj.get_root(doc, comm, False)
        if preprocess:
            root_obj._preprocess(root)
        ref = root.ref['id']
        state._views[ref] = (self, root, doc, comm)
        return root


class FastNumberInput(FastComponent):

    appearance = param.Selector(default='outline', objects=['outline', 'filled'])

    autofocus = param.Boolean(default=False)

    placeholder = param.String(default='Type here')

    step = param.Number(default=0)

    start = param.Number(default=0)

    end = param.Number(default=1)

    value = param.Number(default=0)

    _child_config = {'name': 'template'}

    _template= '<fast-number-field id="fast-number" autofocus="${autofocus}" placeholder="${placeholder}" step="${step}" value="${value}" min="${start}" max="${end}" appearance="${appearance}"></fast-number-field>' # noqa

    _dom_events = {'fast-number': ['change']}


class FastSlider(FastComponent):

    height = param.Integer(default=50)

    step = param.Number(default=0.1)

    start = param.Number(default=0)

    end = param.Number(default=1)

    value = param.Number(default=0)

    _child_config = {'name': 'template'}

    _template = '<fast-slider id="fast-slider" value="${value}" min="${start}" max="${end}" # noqa step="${step}"></fast-slider>'

class FastSelect(FastComponent, SelectBase):

    options = param.ClassSelector(default=[], class_=(dict, list))

    value = param.Parameter()

    _template = """
    <fast-select value="${value}" id="fast-select">
    {% for option in options %}
      <fast-option value="{{ option }}">${option}</fast-option>
    {% endfor %}
    </fast-select>
    """

    def _process_property_change(self, msg):
        msg = super()._process_property_change(msg)
        if 'value' in msg:
            msg['value'] = self._items[msg['value']]
        return msg


class FastButton(FastComponent):

    clicks = param.Integer(default=0, bounds=(0, None))

    value = param.Event()

    _child_config = {'name': 'template'}

    _template = '<fast-button onclick="${_on_click}" appearance="outline" id="fast-button">${name}</fast-button>'

    def _on_click(self, event):
        self.clicks += 1
        self.param.trigger('value')


class FastCheckbox(FastComponent):

    value = param.Boolean(default=True)

    _child_config = {'name': 'template'}

    _template = '<fast-checkbox value="${value}">${name}</fast-checkbox>'


class FastRadioGroup(FastComponent, SelectBase):

    options = param.ClassSelector(default=[], class_=(dict, list))

    value = param.Parameter()

    _child_config = {'options': 'literal'}

    _template = """
    <fast-radio-group value="${value}" id="fast-radio-group">
      <label style="color: var(--neutral-foreground-rest);" slot="label" slot="label">${title}</label>
      {% for option in options %}
      <fast-radio id="fast-radio-{{ loop.index0 }}">${options[{{ loop.index0 }}]}</fast-radio>
      {% endfor %}
      </fast-radio-group>
    """

    def _on_change(self, event):
        self.value = self._items[event.data['value']]

    def _process_children(self, doc, root, model, comm, children):
        children['fast-options'] = self.labels
        return children


class FastToggle(FastComponent):

    value = param.Boolean(default=False)

    _template = '<fast-switch checked="${value}">${name}</fast-switch>'


class FastTextArea(FastComponent):

    value = param.String()

    _template = '<fast-text-area value="${value}"></fast-text-area>'


class FastTextInput(FastComponent):

    placeholder = param.String()

    value = param.String()

    _template = '<fast-text-field placeholder="${placeholder}" value="${value}"></fast-text-field>'


class FastProgress(FastComponent):

    _template = '<fast-progress></fast-progress>'


class FastProgressRing(FastComponent):

    _template = '<fast-progress-ring></fast-progress-ring>'



class FastDivider(FastComponent):

    _template = '<fast-divider></fast-divider>'


class FastCard(FastComponent, ListLike):

    _template = '<fast-card id="fast-card">${objects}</fast-card>'

    def __init__(self, *objects, **params):
        super().__init__(objects=list(objects), **params)


class FastDialog(FastComponent, ListLike):

    hidden = param.Boolean(default=False)

    _template = """
    <fast-dialog id="fast-dialog" hidden=${hidden} style="z-index: 100; --dialog-width: 80%; --dialog-height: 80%">
      <fast-button id="close-button" onclick="${_close}" style="float: right">X</fast-button>
      ${objects}
    </fast-dialog>
    """

    def __init__(self, *objects, **params):
        super().__init__(objects=list(objects), **params)

    def _close(self, event):
        self.hidden = True

class FastTabs(FastComponent, NamedListLike):

    active = param.Integer(default=0, bounds=(0, None), doc="""
        Index of the currently displayed objects.""")

    _activeid = param.String(default='')

    _template = """
    <fast-tabs id="fast-tabs" activeid="${_activeid}">
    {% for obj_name in objects_names %}
      <fast-tab slot="tab" slot="tab">{{ obj_name }}</fast-tab>
    {% endfor %}
    {% for object in objects %}
      <fast-tab-panel id="fast-tab-panel" slot="tabpanel" slot="tabpanel">${object}</fast-tab-panel>
    {% endfor %}
    </fast-tabs>
    """

    @property
    def _child_names(self):
        return {'objects': self._names}

    def __init__(self, *objects, **params):
        NamedListLike.__init__(self, *objects, **params)
        FastComponent.__init__(self, objects=self.objects, **params)

    @param.depends('_activeid', watch=True)
    def _update_active(self):
        self.active = int(self._activeid.split('-')[-1])-1


class FastAccordion(FastComponent, NamedListLike):

    _template = """
    <fast-accordion id="fast-accordion">
      {% for object in objects %}
      <fast-accordion-item slot="item" slot="item">
          <div slot="heading" slot="heading">{{ objects_names[loop.index0] }}</div>
          <svg style="stroke: #e62f63;" width="18" height="18" viewBox="0 0 18 18" fill="none" xmlns="http://www.w3.org/2000/svg" slot="collapsed-icon" slot="collapsed-icon">
            <path d="M15.2222 1H2.77778C1.79594 1 1 1.79594 1 2.77778V15.2222C1 16.2041 1.79594 17 2.77778 17H15.2222C16.2041 17 17 16.2041 17 15.2222V2.77778C17 1.79594 16.2041 1 15.2222 1Z" stroke-linecap="round" stroke-linejoin="round"></path>
            <path d="M9 5.44446V12.5556" stroke-linecap="round" stroke-linejoin="round"></path>
            <path d="M5.44446 9H12.5556" stroke-linecap="round" stroke-linejoin="round"></path>
          </svg>
          <svg style="stroke: #e62f63;" width="18" height="18" viewBox="0 0 18 18" fill="none" xmlns="http://www.w3.org/2000/svg" slot="expanded-icon" slot="expanded-icon">
            <path d="M15.2222 1H2.77778C1.79594 1 1 1.79594 1 2.77778V15.2222C1 16.2041 1.79594 17 2.77778 17H15.2222C16.2041 17 17 16.2041 17 15.2222V2.77778C17 1.79594 16.2041 1 15.2222 1Z" stroke-linecap="round" stroke-linejoin="round"></path>
            <path d="M5.44446 9H12.5556" stroke-linecap="round" stroke-linejoin="round"></path>
          </svg>
          <div id="accordion-content-{{ loop.index0 }}">${objects[{{ loop.index0 }}]}</div>
      </fast-accordion-item>
      {% endfor %}
    </fast-accordion>
    """ # noqa

    def __init__(self, *objects, **params):
        NamedListLike.__init__(self, *objects, **params)
        FastComponent.__init__(self, objects=self.objects, **params)

    @property
    def _child_names(self):
        return {'objects': self._names}



import panel as pn
import param

from panel.reactive import ReactiveHTML

from lumen.sources import Source
from lumen.views import View
from .base import WizardItem
from .gallery import GalleryItem, Gallery
from .sources import SourceGallery, ASSETS_DIR
from .state import state


class ViewsEditor(WizardItem):
    """
    Declare the views for your dashboard.
    """

    spec = param.List(default=[], precedence=-1)

    source_gallery = param.ClassSelector(class_=SourceGallery, precedence=-1)

    source = param.String()

    view_type = param.Selector()

    sources = param.List(label='Select a source')

    tables = param.List(label='Select a table')

    table = param.String()

    views = param.List()

    _template = """
    <span style="font-size: 2em">View Editor</span>
    <p>{{ __doc__ }}</p>
    <fast-divider></fast-divider>
    <div style="display: flex; height: 100%;">
      <form role="form" style="flex: 20%; max-width: 250px; line-height: 2em;">
        <div style="display: grid;">
          <label for="sources-${id}"><b>{{ param.sources.label }}</b></label>
          <fast-select id="source" style="max-width: 250px; min-width: 150px;" value="${source}">
          {% for src in sources %}
            <fast-option value="{{ src }}">{{ src.title() }}</fast-option>
          {% endfor %}
          </fast-select>
          <fast-tooltip anchor="sources-${id}">{{ param.sources.doc }}</fast-tooltip>
        </div>
        <div style="display: grid;">
          <label for="table-${id}"><b>{{ param.tables.label }}</b></label>
          <fast-select id="tables" style="max-width: 250px; min-width: 150px;" value="${table}">
          {% for tbl in tables %}
            <fast-option value="{{ tbl }}">{{ tbl }}</fast-option>
          {% endfor %}
          </fast-select>
          <fast-tooltip anchor="tables-${id}">{{ param.tables.doc }}</fast-tooltip>
        </div>
        <div style="display: flex;">
          <div style="display: grid; flex: auto;">
            <label for="view-select-${id}"><b>{{ param.view_type.label }}</b></label>
            <fast-select id="view-select" style="min-width: 150px;" value="${view_type}">
            {% for vtype in param.view_type.objects %}
              <fast-option value="{{ vtype }}">{{ vtype.title() }}</fast-option>
            {% endfor %}
            </fast-select>
            <fast-tooltip anchor="view-select-${id}">{{ param.view_type.doc }}</fast-tooltip>
          </div>
          <fast-button id="submit" appearance="accent" style="margin-top: auto; margin-left: 1em; width: 20px;" onclick="${_add_view}">
            <b style="font-size: 2em;">+</b>
          </fast-button>
        </div>
      </form>
      <div style="flex: auto; margin-left: 1em; overflow-y: auto;">
        {% for view in views %}
        <div id="view-container">${view}</div>
        <fast-divider></faster-divider>
        {% endfor %}
      </div>
    </div>
    """

    def __init__(self, **params):
        super().__init__(**params)
        views = param.concrete_descendents(View)
        self.param.view_type.objects = [source.view_type for source in views.values()]
        self._source = None
        state.sources.param.watch(self._update_sources, 'sources')

    @param.depends('source', watch=True)
    def _update_tables(self):
        source = state.sources.sources[self.source]
        spec = dict(source.spec, cache_dir=None)
        spec.pop('filters', None)
        self._source = Source.from_spec(spec)
        schema = self._source.get_schema()
        self.tables = list(schema)

    def _add_view(self, event):
        editor = ViewEditor(
            view_type=self.view_type, source=self.source,
            table=self.table, source_obj=self._source,
            spec={'table': self.table, 'type': self.view_type}
        )
        editor.param.watch(self._remove_view, 'remove')
        self.views.append(editor)
        self.param.trigger('views')

    def _update_sources(self, event):
        self.sources = list(event.new)
        if not self.source and self.sources:
            self.source = self.sources[0]

    def _remove_view(self, event):
        self.views.remove(event.obj)
        self.param.trigger('views')


class ViewEditor(ReactiveHTML):

    source = param.Parameter(precedence=-1)

    source_obj = param.Parameter(precedence=-1)

    remove = param.Boolean(default=False)

    spec = param.Dict(default={})

    sizing_mode = param.String(default='stretch_both')

    table = param.String()

    view_type = param.String(default='None')
    
    view = param.Parameter()

    _template = """
    <span style="font-size: 2em">{{ view_type.title() }} Editor</span>
    <p>Configure the view.</p>
    <fast-divider></fast-divider>
    <div id='view'>${view}</div>
    """

    def __new__(cls, **params):
        if cls is not ViewEditor:
            return super().__new__(cls)
        editors = param.concrete_descendents(cls)
        for editor in editors.values():
            if editor.view_type == params['view_type']:
                return super().__new__(editor)
        return super().__new__(cls)

    @property
    def description(self):
        return f'A {self.view_type} view of the {self.table} table on the {self.source_obj.name}'

    @property
    def thumbnail(self):
        return ASSETS_DIR / 'view.png'

    def _remove(self, event):
        self.remove = True

    def _update_spec(self, *events):
        for event in events:
            if not isinstance(event.new, param.Parameterized):
                self.spec[event.name] = event.new


class ViewGalleryItem(GalleryItem):

    editor = param.ClassSelector(class_=ViewEditor, precedence=-1)

    _template = """
    <span style="font-size: 1.2em; font-weight: bold;">{{ name }}</p>
    <fast-switch id="selected" checked=${selected} style="float: right"></fast-switch>
    <div id="details" style="margin: 1em 0;">
      ${view}
    </div>
    <p style="height: 4em; max-width: 320px;">{{ description }}</p>
    <fast-button id="edit-button" style="width: 320px;" onclick="${_open_modal}">Edit</fast-button>
    """

    def __init__(self, **params):
        super().__init__(**params)
        self.thumbnail = self.editor.thumbnail
        self.description = self.editor.description
        self.view = pn.pane.PNG(self.thumbnail, height=200, align='center')
        self._modal_content = [self.editor]

    @param.depends('selected', watch=True)
    def _add_spec(self):
        views = state.spec['views']
        if self.selected:
            views[self.name] = self.spec
        elif self.name in views:
            del views[self.name]


class ViewGallery(WizardItem, Gallery):
    """
    Select the views to add to your dashboard specification.
    """

    path = param.Foldername()

    spec = param.List()

    views = param.List(default=[], doc="The list of views added to the dashboard.", precedence=-1)

    _template = """
    <span style="font-size: 1.5em">Views</span>
    <fast-divider></fast-divider>
    <span style="font-size: 1.2em; font-weight: bold;">{{ __doc__ }}</p>
    <div id="items" style="margin: 1em 0; display: flex; flex-wrap: wrap; gap: 1em;">
    {% for item in items %}
      <fast-card id="view-container" class="gallery-item" style="width: 350px; height: 400px;">
        ${item}
      </fast-card>
    {% endfor %}
      <fast-card id="view-container-new" class="gallery-item" style="height: 400px; width: 350px; padding: 1em;">
        <div style="display: grid;">
          <span style="font-size: 1.25em; font-weight: bold;">Add new view</span>
          <i id="add-button" onclick="${_create_new}" class="fa fa-plus" style="font-size: 14em; margin: 0.2em auto;" aria-hidden="true"></i>
        </div>
      </fast-card>
    </div>
    """

    _editor_type = ViewEditor

    _gallery_item = ViewGalleryItem

    def __init__(self, **params):
        super().__init__(**params)
        self._editor = ViewsEditor(spec=self.spec, margin=10)
        self._save_button = pn.widgets.Button(name='Save view')
        self._save_button.on_click(self._save_view)

    def _create_new(self, event):
        if state.modal.objects == [self._editor, self._save_button]:
            state.template.open_modal()
            return
        state.modal.loading = True
        state.template.open_modal()
        state.modal[:] = [self._editor, self._save_button]
        state.modal.loading = False

    def _save_view(self, event):
        for view in self._editor.views:
            self.items[view.name] = ViewGalleryItem(
                name=view.name, spec=view.spec, selected=True, editor=view,
                thumbnail=view.thumbnail
            )
            self.views.append(view)
        self.param.trigger('views')    
        self.param.trigger('items')
        self._editor.views = []
        state.template.close_modal()

    @param.depends('spec', watch=True)
    def _update_params(self):
        for name, spec in self.spec.items():
            self.views = editor = ViewEditor(name=name, spec=spec)
            self.items[name] = ViewGalleryItem(
                name=name, spec=spec, selected=True, editor=editor,
                thumbnail=editor.thumbnail
            )
        self.param.trigger('views')
        self.param.trigger('items')


class TableViewEditor(ViewEditor):

    view_type = param.String(default='table')

    def __init__(self, **params):
        super().__init__(**{k: v for k, v in params.items() if k in self.param})
        kwargs = dict(self.spec)
        df = self.source_obj.get(kwargs.pop('table'))
        if 'sizing_mode' not in kwargs:
            kwargs['sizing_mode'] = 'stretch_both'
        self.tabulator = pn.widgets.Tabulator(
            df, pagination='remote', page_size=15, **kwargs
        )
        controls = ['theme', 'layout', 'page_size']
        self.view = pn.Row(
            self.tabulator.controls(controls, width=300),
            self.tabulator,
            sizing_mode='stretch_width'
        )
        self.tabulator.param.watch(self._update_spec, controls)

    @property
    def thumbnail(self):
        return ASSETS_DIR / 'tabulator.png'


class PerspectiveViewEditor(ViewEditor):

    view_type = param.String(default='perspective')

    _defaults = dict(sizing_mode='stretch_both', min_height=500, theme='material-dark')

    def __init__(self, **params):
        spec = params.get('spec', {})
        spec['height'] = 500 # remove
        params['view'] = view = pn.pane.Perspective(**dict(self._defaults, **spec))
        super().__init__(**{k: v for k, v in params.items() if k in self.param})
        view.object = self.source_obj.get(self.table)
        view.param.watch(self._update_spec, list(view.param))

    @property
    def thumbnail(self):
        return ASSETS_DIR / 'perspective.png'


class hvPlotViewEditor(ViewEditor):

    selection_group = param.String(default=None, allow_None=True, precedence=-1)

    view_type = param.String(default='hvplot')

    def __init__(self, **params):
        import hvplot.pandas # noqa
        from hvplot.ui import hvPlotExplorer
        super().__init__(**params)
        kwargs = dict(self.spec)
        df = self.source_obj.get(kwargs.pop('table'))
        self.view = hvPlotExplorer(df, **kwargs)
        self.view.param.watch(self._update_spec, list(self.view.param))
        self.view.param.trigger('kind', 'x', 'y')

    @property
    def thumbnail(self):
        return ASSETS_DIR / 'hvplot.png'

    def _update_spec(self, *events):
        for event in events:
            p = event.name
            p = 'y' if p == 'y_multi' else p
            if not isinstance(event.new, param.Parameterized):
                self.spec[p] = event.new



from weakref import WeakKeyDictionary

import panel as pn

from .config import config


class _session_state:
    """
    The session_state object holds both global and session specific
    state making it easy to manage access to Sources and Filters
    across components.
    """

    global_sources = {}

    global_filters = {}

    spec = {}

    _loading = WeakKeyDictionary() if pn.state.curdoc else {}

    _sources = WeakKeyDictionary() if pn.state.curdoc else {}

    _filters = WeakKeyDictionary() if pn.state.curdoc else {}

    @property
    def filters(self):
        from .filters import Filter
        if pn.state.curdoc not in self._filters:
            self._filters[pn.state.curdoc] = {
                source: {
                    name: Filter.from_spec(spec, schema)
                    for name, (spec, schema) in filters.items()
                }
                for source, filters in self.global_filters.items()
            }
        return self._filters[pn.state.curdoc]

    @property
    def sources(self):
        if pn.state.curdoc not in self._sources:
            self._sources[pn.state.curdoc] = dict(self.global_sources)
        return self._sources[pn.state.curdoc]

    @property
    def loading_msg(self):
        return self._loading.get(pn.state.curdoc)

    @loading_msg.setter
    def loading_msg(self, loading_msg):
        self._loading[pn.state.curdoc] = loading_msg

    @property
    def authorized(self):
        if pn.state.user_info is None and self.spec.get('auth'):
            return config.dev
        authorized = True
        for k, value in self.spec.get('auth', {}).items():
            if not isinstance(value, list): value = [value]
            if k in pn.state.user_info:
                user_value = pn.state.user_info[k]
                if not isinstance(user_value, list):
                    user_value = [user_value]
                authorized &= any(uv == v for v in value for uv in user_value)
        return authorized

    def load_global_sources(self, clear_cache=True):
        """
        Loads global sources shared across all targets.
        """
        from .sources import Source
        for name, source_spec in self.spec.get('sources', {}).items():
            if not source_spec.get('shared'):
                continue
            source_spec = dict(source_spec)
            filter_specs = source_spec.pop('filters', {})
            source = self.global_sources.get(name)
            source_type = Source._get_type(source_spec['type'])
            if source is not None:
                # Check if old source is the same as the new source
                params = {
                    k: v for k, v in source.param.get_param_values()
                    if k != 'name'
                }
                new_spec = dict(source_spec, **{
                    k: v for k, v in source_type.param.get_param_values()
                    if k != 'name'
                })
                reinitialize = params == new_spec
            else:
                reinitialize = True
            if self.loading_msg is not None:
                self.loading_msg.object = f'Loading {name} source...'
            if reinitialize:
                source_spec['name'] = name
                self.global_sources[name] = source = Source.from_spec(source_spec)
            if source.cache_dir and clear_cache:
                source.clear_cache()
            schema = source.get_schema()
            self.global_filters[name] = {
                fname: (filter_spec, schema)
                for fname, filter_spec in filter_specs.items()
            }

    def load_source(self, name, source_spec):
        from .filters import Filter
        from .sources import Source
        source_spec = dict(source_spec)
        filter_specs = source_spec.pop('filters', None)
        if self.loading_msg:
            self.loading_msg.object = f'Loading {name} source...'
        if name in self.sources:
            source = self.sources[name]
        else:
            source_spec['name'] = name
            source = Source.from_spec(source_spec)
        self.sources[name] = source
        if not filter_specs:
            return source
        self.filters[name] = filters = dict(self.filters.get(name, {}))
        schema = source.get_schema()
        for fname, filter_spec in filter_specs.items():
            if fname not in filters:
                filters[fname] = Filter.from_spec(filter_spec, schema)
        return source

    def resolve_views(self):
        from .views import View
        exts = []
        for target in self.spec.get('targets', []):
            views = target.get('views', [])
            if isinstance(views, dict):
                views = list(views.values())
            for view in views:
                view_type = View._get_type(view.get('type'))
                if view_type and view_type._extension:
                    exts.append(view_type._extension)
        for ext in exts:
            __import__(pn.extension._imports[ext])


state = _session_state()



import importlib
import os
import sys

import param as param
import panel as pn

from panel.widgets.indicators import Indicator
from panel.template.base import BasicTemplate
from panel.template import DefaultTheme, DarkTheme


_INDICATORS = {k.lower(): v for k, v in param.concrete_descendents(Indicator).items()}

_LAYOUTS = {
    'accordion': pn.Accordion,
    'column'   : pn.Column,
    'grid'     : pn.GridBox,
    'row'      : pn.Row,
    'tabs'     : pn.Tabs
}

try:
    _DEFAULT_LAYOUT = _LAYOUTS['flex'] = pn.FlexBox
except Exception:
    _DEFAULT_LAYOUT = pn.GridBox

_TEMPLATES = {
    k[:-8].lower(): v for k, v in param.concrete_descendents(BasicTemplate).items()
}

_THEMES = {'default': DefaultTheme, 'dark': DarkTheme}



class _config(param.Parameterized):
    """
    Stores shared configuration for the entire Lumen application.
    """

    _root = param.String(default=None, doc="""
      Root directory relative to which data and modules are loaded.""")

    _dev = param.Boolean(default=None, doc="""
      Whether Lumen should run in development mode.""")

    template_vars = param.Dict(default={}, doc="""
      Template variables which may be referenced in a dashboard yaml
      specification.""")

    yamls = param.List(default=[], doc="""
      List of yaml files currently being served.""")

    _modules = {}

    @property
    def root(self):
        if self._root:
            return self._root
        return os.getcwd()

    def load_local_modules(self):
        """
        Loads local modules containing custom components and templates.
        """
        for imp in ('filters', 'sources', 'transforms', 'template', 'views'):
            path = os.path.join(self.root, imp+'.py')
            if not os.path.isfile(path):
                continue
            spec = importlib.util.spec_from_file_location(f"local_lumen.{imp}", path)
            if path in self._modules:
                continue
            self._modules[path] = module = importlib.util.module_from_spec(spec)
            spec.loader.exec_module(module)
        templates = param.concrete_descendents(BasicTemplate)
        _TEMPLATES.update({
            k[:-8].lower(): v for k, v in templates.items()
        })

    @property
    def dev(self):
        """
        Whether the application was launched in development mode.
        """
        if self._dev is not None:
            return self._dev
        return '--dev' in sys.argv or '--autoreload' in sys.argv

    @dev.setter
    def dev(self, dev):
        self._dev = dev




config = _config()



import datetime as dt
import importlib
import re
import os
import sys
import subprocess

from jinja2 import Environment, Undefined, DebugUndefined
from pandas.core.dtypes.dtypes import CategoricalDtype
from panel import state


def get_dataframe_schema(df, columns=None):
    """
    Returns a JSON schema optionally filtered by a subset of the columns.

    Parameters
    ----------
    df : pandas.DataFrame or dask.DataFrame
        The DataFrame to describe with the schema
    columns: list(str) or None
        List of columns to include in schema

    Returns
    -------
    dict
        The JSON schema describing the DataFrame
    """
    if 'dask.dataframe' in sys.modules:
        import dask.dataframe as dd
        is_dask = isinstance(df, dd.DataFrame)
    else:
        is_dask = False

    if columns is None:
        columns = list(df.columns)

    schema = {'type': 'array', 'items': {'type': 'object', 'properties': {}}}
    properties = schema['items']['properties']
    for name in columns:
        dtype = df.dtypes[name]
        column = df[name]
        if dtype.kind in 'uifM':
            vmin, vmax = column.min(), column.max()
            if is_dask:
                vmin, vmax = dd.compute(vmin, vmax)
            if dtype.kind == 'M':
                kind = 'string'
                vmin, vmax = vmin.isoformat(), vmax.isoformat()
            else:
                if dtype.kind == 'f':
                    cast = float
                    kind = 'number'
                else:
                    cast = int
                    kind = 'integer'
                vmin, vmax = cast(vmin), cast(vmax)
            properties[name] = {
                'type': kind,
                'inclusiveMinimum': vmin,
                'inclusiveMaximum': vmax
            }
            if dtype.kind == 'M':
                properties[name]['format'] = 'datetime'
        elif dtype.kind == 'b':
            properties[name] = {'type': 'boolean'}
        elif dtype.kind == 'O':
            if isinstance(dtype, CategoricalDtype) and len(dtype.categories):
                cats = list(dtype.categories)
            else:
                try:
                    cats = column.unique()
                    if is_dask:
                        cats = cats.compute()
                except Exception:
                    cats = []
                cats = list(cats)
            properties[name] = {'type': 'string', 'enum': cats}
    return schema

_period_regex = re.compile(r'((?P<weeks>\d+?)w)?((?P<days>\d+?)d)?((?P<hours>\d+?)h)?((?P<minutes>\d+?)m)?((?P<seconds>\d+?)s)?')


def parse_timedelta(time_str):
    parts = _period_regex.match(time_str)
    if not parts:
        return
    parts = parts.groupdict()
    time_params = {}
    for (name, p) in parts.items():
        if p:
            time_params[name] = int(p)
    return dt.timedelta(**time_params)


def _j_getenv(x):
    if isinstance(x, Undefined):
        x = x._undefined_name
    return os.getenv(x, '')

def _j_getshell(x):
    if isinstance(x, Undefined):
        x = x._undefined_name
    try:
        return subprocess.check_output(x).decode()
    except (IOError, OSError):
        return ""

def _j_getheaders(x):
    if isinstance(x, Undefined):
        x = x._undefined_name
    return state.headers.get(x, '')

def _j_getcookies(x):
    if isinstance(x, Undefined):
        x = x._undefined_name
    return state.cookies.get(x, '')

def _j_getoauth(x):
    if isinstance(x, Undefined):
        x = x._undefined_name
    if state.user_info is None:
        return ''
    return state.user_info.get(x, '')

def expand_spec(pars, context={}, getenv=True, getshell=True, getheaders=True,
                getcookies=True, getoauth=True):
    """
    Render variables in context into the set of parameters with jinja2.

    For variables that are not strings, nothing happens.

    Parameters
    ----------
    pars: dict
        values are strings containing some jinja2 controls
    context: dict
        values to use while rendering

    Returns
    -------
    dict with the same keys as ``pars``, but updated values
    """
    if isinstance(pars, dict):
        return {k: expand_spec(
            v, context, getenv, getshell, getheaders, getcookies, getoauth
        ) for k, v in pars.items()}
    elif isinstance(pars, (list, tuple, set)):
        return type(pars)(expand_spec(
            v, context, getenv, getshell, getheaders, getcookies, getoauth
        ) for v in pars)
    elif isinstance(pars, str):
        jinja = Environment(undefined=DebugUndefined)
        if getenv:
            jinja.globals['env'] = _j_getenv
        if getshell:
            jinja.globals['shell'] = _j_getshell
        if getheaders:
            jinja.globals['header'] = _j_getheaders
        if getcookies:
            jinja.globals['cookie'] = _j_getcookies
        if getoauth:
            jinja.globals['oauth'] = _j_getoauth
        return jinja.from_string(pars).render(context)
    else:
        # no expansion
        return pars


def merge_schemas(schema, old_schema):
    """
    Merges two JSON schemas on a column.
    """
    if old_schema is None:
        return schema
    elif schema['type'] != old_schema['type']:
        return old_schema
    elif 'enum' in schema and 'enum' in old_schema:
        merged_enum = list(old_schema['enum'])
        for enum in schema['enum']:
            if enum not in merged_enum:
                merged_enum.append(enum)
        return dict(old_schema, enum=merged_enum)
    elif 'inclusiveMinimum' in schema and 'inclusiveMinimum' in old_schema:
        merged_min = min(schema['inclusiveMinimum'], old_schema['inclusiveMinimum'])
        merged_max = min(schema['inclusiveMaximum'], old_schema['inclusiveMaximum'])
        return dict(old_schema, inclusiveMinimum=merged_min, inclusiveMaximum=merged_max)


def resolve_module_reference(reference, component_type):
    cls_name = component_type.__name__
    *modules, ctype = reference.split('.')
    module = '.'.join(modules)
    try:
        module = importlib.import_module(module)
    except Exception:
        raise ValueError(f"{cls_name} type '{reference}' module could "
                         f"not be resolved. Ensure explicit {cls_name} "
                         "type references a valid module.")
    if not hasattr(module, ctype):
        raise ValueError(f"Source type '{reference}' could not be "
                         f"resolved. Module '{module}' has no member "
                         f"{ctype}.")
    component = getattr(module, ctype)
    if not issubclass(component, component_type):
        raise ValueError(f"{cls_name} type '{reference}' did not resolve "
                         "to a {cls_name} subclass.")
    return component
